from .quantity import Quantity
from .temperature import Temperature
from .energy import Energy


class GasConstant(Quantity):

    si = 8.314
    JmolK = si
    m3PaKmol = si
    kJmolK = 0.008314
    LbarKmol = 0.08314

    @classmethod
    def __mul__(cls, other):
        if isinstance(other, Temperature):
            si = cls.si * other.si
            return Energy(si=si)
        return NotImplemented

    @classmethod
    def __rmul__(cls, other):
        if isinstance(other, Temperature):
            si = other.si * cls.si
            return Energy(si=si)
        return NotImplemented



from .species import Species
import catrxneng.quantities as quant
import catrxneng.utils.equations as eqn


class H2(Species):
    CLASS = "diatomic_gas"
    H_ATOMS = 2
    MOL_WEIGHT = 2
    HF_298_GAS = quant.Energy(kJmol=0)
    S_298_GAS = quant.Entropy(JmolK=130.68)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 1000,
            "phase": "gas",
            "A": 33.066178,
            "B": -11.363417,
            "C": 11.432816,
            "D": -2.772874,
            "E": -0.158558,
            "F": -9.980797,
            "G": 172.707974,
            "H": 0,
        }
    ]

    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def S_gas(cls, T):
        return eqn.S_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def Cp_gas(cls, T):
        return eqn.Cp_shomate(T, cls.NIST_THERMO_PARAMS[0])



import pandas as pd, numpy as np

from .step import Step
from catrxneng import utils
import catrxneng.quantities as quant


class Co2FtsStep(Step):

    def __init__(self, step_name, step_num, start, end):
        super().__init__(step_name, step_num, start, end)

    def attach_reactor(self, reactor_class, kinetic_model, T, p0, whsv, mcat):
        self.reactor = reactor_class(kinetic_model, T, p0, whsv, "co2", mcat=mcat)

    def simulate(self, dt_sec, std_dev=None):
        if std_dev is None:
            std_dev = {"temp": 0.3, "pressure": 0.1, "mfc": 0.5, "gc": 0.005}
        df = pd.DataFrame()
        df["timestamp"] = np.arange(self.start.UET, self.end.UET + dt_sec, dt_sec)
        num_points = df["timestamp"].size
        df["step_name"] = self.step_name
        df["step_num"] = self.step_num
        self.reactor.solve()
        df["bed_temp"] = self.reactor.T.C + np.random.normal(
            loc=0, scale=std_dev["temp"], size=num_points
        )
        df["pressure"] = self.reactor.P.bar + np.random.normal(
            loc=0, scale=std_dev["pressure"], size=num_points
        )
        df["n2_mfc"] = self.reactor.F0["inert"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )
        df["co2_mfc"] = self.reactor.F0["co2"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )
        df["h2_mfc"] = self.reactor.F0["h2"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )

        ch4_sel = 0.3
        olefins_sel = 0.4
        paraffins_sel = 1 - ch4_sel - olefins_sel
        F_hydrocarbons = (
            self.reactor.F0["co2"]
            - self.reactor.F["co2"][-1]
            - self.reactor.F["co"][-1]
        )

        F_out = self.reactor.F[:, -1]
        F_out.keys[F_out.keys.index("inert")] = "n2"
        F_out["ch4"] = F_hydrocarbons * ch4_sel
        F_out["c2h4"] = F_hydrocarbons * olefins_sel / 4 / 2
        F_out["c3h6"] = F_hydrocarbons * olefins_sel / 4 / 3
        F_out["c4h8"] = F_hydrocarbons * olefins_sel / 4 / 4
        F_out["c5h10"] = F_hydrocarbons * olefins_sel / 4 / 5
        F_out["c2h6"] = F_hydrocarbons * paraffins_sel / 4 / 2
        F_out["c3h8"] = F_hydrocarbons * paraffins_sel / 4 / 3
        F_out["c4h10"] = F_hydrocarbons * paraffins_sel / 4 / 4
        F_out["c5h12"] = F_hydrocarbons * paraffins_sel / 4 / 5

        y_out = F_out / self.reactor.Ft[-1]

        for component in y_out.keys:
            df[f"{component}_gc_conc"] = y_out[component].pct * np.random.normal(
                loc=1, scale=std_dev["gc"], size=num_points
            )

        delta_co2 = self.reactor.F0["co2"] - F_out["co2"]
        self.co2_conv = utils.divide(delta_co2, self.reactor.F0["co2"])
        conf = utils.getconf("co2_fts", "components")
        carbon_out = np.sum(
            [F_out[comp_id].si * conf[comp_id].get("C_ATOMS", 0) for comp_id in conf]
        )
        carbon_out = quant.MolarFlowRate(si=carbon_out)
        self.carbon_bal = carbon_out / self.reactor.F0["co2"]

        self.time_series_data = df



import numpy as np

import catrxneng.species as species
from catrxneng.reactions import RWGS, FTS
import catrxneng.quantities as quant


class Brubach2022:
    LIMITING_REACTANT = "co2"
    T_REF = quant.Temperature(C=300)
    COMPONENTS = {
        "co2": species.CO2,
        "h2": species.H2,
        "co": species.CO,
        "h2o": species.H2O,
        "c2h4": species.C2H4,
        "inert": species.Inert,
    }
    REACTIONS = {
        "rwgs": RWGS,
        "fts": FTS,
    }
    ORDER = np.array([1.5, 2.0])
    STOICH_COEFF = np.array(
        [
            [-1, -1, 1, 1, 0, 0],
            [0, -4, 2, 2, 1, 0],
        ]
    )
    a_rwgs = quant.Dimensionless(si=16.3)
    a_ft = quant.Dimensionless(si=9.07)
    b_ft = quant.InversePressure(inv_bar=2.44)

    def __init__(self, T=None, kref=None, Ea=None):
        self.kref = kref
        if self.kref is None:
            self.kref = np.array([8.13e-2, 6.39e-2])
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = np.array([115, 67.8])  # kJ/mol
        super().__init__(T)

    def k_rwgs(self, T):
        return rate_const(T=T, Ea=self.Ea_rwgs, kref=self.kref_rwgs, Tref=self.Tref)

    def k_ft(self, T):
        return rate_const(T=T, Ea=self.Ea_ft, kref=self.kref_ft, Tref=self.Tref)

    def r_rwgs(self, T, p):
        fwd = p["co2"] * p["h2"] ** 0.5
        rev = p["co"] * p["h2o"] / (RWGS(T).Keq * p["h2"] ** 0.5)
        num = self.k_rwgs(T) * (fwd - rev)
        denom = (1 + self.a_rwgs * p["h2o"] / p["h2"]) ** 2
        rate = num / denom
        return quant.ReactionRate(si=rate.si)

    def r_ft(self, T, p):
        num = self.k_ft(T) * p["h2"] * p["co"]
        denom = (1 + self.a_ft * p["h2o"] / p[1] + self.b_ft * p["co"]) ** 2
        rate = num / denom
        return quant.ReactionRate(si=rate.si)

    @property
    def reaction_rate(self):
        return {
            "co2": lambda T, p: -self.r_rwgs(T, p),
            "h2": lambda T, p: -self.r_rwgs(T, p) - 2 * self.r_ft(T, p),
            "co": lambda T, p: self.r_rwgs(T, p) - self.r_ft(T, p),
            "h2o": lambda T, p: self.r_rwgs(T, p) + self.r_ft(T, p),
            "inert": lambda T, p: quant.ReactionRate(si=0),
        }



import numpy as np
from typing import Any

import catrxneng.quantities as quant


def Hf_shomate(T, params):
    from catrxneng.quantities import Energy

    t = T.K / 1000
    dHf = (
        params["A"] * t
        + params["B"] * t**2 / 2
        + params["C"] * t**3 / 3
        + params["D"] * t**4 / 4
        - params["E"] / t
        + params["F"]
        # - params["H"]
    )  # kJ/mol
    return Energy(kJmol=dHf)


def S_shomate(T, params):
    from catrxneng.quantities import Entropy

    t = T.K / 1000
    S = (
        params["A"] * np.log(t)
        + params["B"] * t
        + params["C"] * t**2 / 2
        + params["D"] * t**3 / 3
        - params["E"] / (2 * t**2)
        + params["G"]
    )  # J/mol/K
    return Entropy(JmolK=S)


def Cp_shomate(T, params):
    from catrxneng.quantities import HeatCapacity

    t = T.K / 1000
    Cp = (
        params["A"]
        + params["B"] * t
        + params["C"] * t**2
        + params["D"] * t**3
        + params["E"] / (t**2)
    )
    return HeatCapacity(JmolK=Cp)


def vant_hoff_eqn(
    x_ref: Any, dH: quant.Energy, T: quant.Temperature, Tref: quant.Temperature
):
    return x_ref * np.exp(dH / quant.R * (1 / Tref - 1 / T))



from catrxneng import utils
from .species import Species
from ..quantities import *


class N2(Species):
    CLASS = "diatomic_gas"
    N_ATOMS = 2
    MOL_WEIGHT = 28
    HF_298_GAS = Energy(kJmol=0)
    S_298_GAS = Entropy(JmolK=191.61)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 100,
            "max_temp_K": 500,
            "phase": "gas",
            "A": 28.98641,
            "B": 1.853978,
            "C": -9.647459,
            "D": 16.63537,
            "E": 0.000117,
            "F": -8.671914,
            "G": 226.4168,
            "H": 0.0,
        },
        {
            "min_temp_K": 500,
            "max_temp_K": 2000,
            "phase": "gas",
            "A": 19.50583,
            "B": 19.88705,
            "C": -8.598535,
            "D": 1.369784,
            "E": 0.527601,
            "F": -4.935202,
            "G": 212.3900,
            "H": 0.0,
        },
    ]

    @classmethod
    def Hf_gas(cls, T):
        thermo_params = cls._get_thermo_params(T)
        return eqn.Hf_shomate(T, thermo_params)

    @classmethod
    def S_gas(cls, T):
        thermo_params = cls._get_thermo_params(T)
        return eqn.S_shomate(T, thermo_params)

    @classmethod
    def Cp_gas(cls, T):
        thermo_params = cls._get_thermo_params(T)
        return eqn.Cp_shomate(T, thermo_params)



from .species import Species
import catrxneng.utils.equations as eqn
import catrxneng.quantities as quant


class CH4(Species):
    CLASS = "alkane"
    C_ATOMS = 1
    H_ATOMS = 4
    MOL_WEIGHT = 16
    HF_298_GAS = quant.Energy(kJmol=-74.6)
    S_298_GAS = quant.Entropy(JmolK=186.3)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 1300,
            "phase": "gas",
            "A": -0.703029,
            "B": 108.4773,
            "C": -42.52157,
            "D": 5.862788,
            "E": 0.678565,
            "F": -76.84376,
            "G": 158.7163,
            "H": -74.87310,
        }
    ]

    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def S_gas(cls, T):
        return eqn.S_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def Cp_gas(cls, T):
        return eqn.Cp_shomate(T, cls.NIST_THERMO_PARAMS[0])



import numpy as np

from ..kinetic_model import KineticModel
import catrxneng.utils.equations as eqn
from catrxneng.reactions import RWGS, Co2ToMeoh, Sabatier
import catrxneng.quantities as quant


class Ghosh2022Full(KineticModel):
    def __init__(self, limiting_reactant="co2", T=None):
        self.Tref = quant.Temperature(C=320)
        self.components = [
            "co2",
            "h2",
            "ch3oh",
            "h2o",
            "co",
            "ch4",
            "c2h4",
            "c3h6",
            "c4h8",
            "c2h6",
            "c3h8",
            "c4h10",
            "c58",
            "c9+",
            "inert",
        ]
        self.f8 = {
            "c2h4": 5e-2,
            "c3h6": 1.0,
            "c4h8": 2.6e-1,
        }
        self.f9 = {
            "c2h4": 5.5e-1,
            "c3h6": 1.0,
            "c4h8": 4.5e-1,
        }
        self.kref = np.array(
            [
                8.8e-4,
                2.6e-3,
                1.4e-4,
                5.9e-1,
                7e-2,
                6e-1,
                5.9e-2,
                8e-2,
                8.3e-1,
                6e-2,
                8.2e-1,
            ]
        )
        self.Ea = np.array(
            [35.7, 54.5, 42.5, 70.0, 16.0, 17.0, 69.0, 109.0, 150.0, 170.0, 25.0]
        )
        self.order = np.array([2, 1.5, 1, 1, 2, 2, 2, 2, 1, 1, 2])
        self.comp_idx = {comp: i for i, comp in enumerate(self.components)}
        self.reactions = {
            "co2_to_meoh": Co2ToMeoh(),
            "rwgs": RWGS(),
            "sabatier": Sabatier(),
        }
        super().__init__(limiting_reactant, T)

    def compute_temp_dependent_constants(self):
        self.K_co2 = eqn.vant_hoff_eqn(
            6.7e-1, quant.Energy(kJmol=-2.6e1), self.T, self.Tref
        ).si
        self.K_h2 = eqn.vant_hoff_eqn(
            7e-1, quant.Energy(kJmol=-1.2e1), self.T, self.Tref
        ).si
        self.K_ch3oh = eqn.vant_hoff_eqn(
            1.3e1, quant.Energy(kJmol=-2e-1), self.T, self.Tref
        ).si
        self.K_co2_to_meoh = self.reactions["co2_to_meoh"].Keq(self.T)
        self.K_rwgs = self.reactions["rwgs"].Keq(self.T)
        self.K_sabatier = self.reactions["sabatier"].Keq(self.T)
        self.Keq = np.array([self.K_co2_to_meoh, self.K_rwgs, self.K_sabatier])
        self.k = np.array(
            [
                quant.RateConstant(
                    molskgcatbar=kref,
                    Ea=quant.Energy(kJmol=Ea),
                    Tref=self.Tref,
                    order=order,
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.order)
            ]
        )

    def rate_equations(self, p_array):
        p_co2 = p_array[self.comp_idx["co2"]]
        p_h2 = p_array[self.comp_idx["h2"]]
        p_ch3oh = p_array[self.comp_idx["ch3oh"]]
        p_h2o = p_array[self.comp_idx["h2o"]]
        p_co = p_array[self.comp_idx["co"]]
        p_ch4 = p_array[self.comp_idx["ch4"]]
        p_c2h4 = p_array[self.comp_idx["c2h4"]]
        p_c3h6 = p_array[self.comp_idx["c3h6"]]
        p_c4h8 = p_array[self.comp_idx["c4h8"]]
        p_c58 = p_array[self.comp_idx["c58"]]

        n = 1

        # CO2 hydrogenation reactions
        inhib = (1 + self.K_co2 * p_co2 + np.sqrt(self.K_h2 * p_h2)) ** 2
        fwd = p_co2 * p_h2**3
        rev = p_ch3oh * p_h2o / self.K_co2_to_meoh
        numerator = self.k[1 - n] * (fwd - rev) / (p_h2**2)
        r1 = numerator / inhib

        fwd = p_co2 * p_h2
        rev = p_co * p_h2o / self.K_rwgs
        numerator = self.k[2 - n] * (fwd - rev) / np.sqrt(p_h2)
        r2 = numerator / inhib

        numerator = p_ch4 * p_h2o**2
        denom = p_co2 * p_h2**4 * self.K_sabatier
        frac = (1 - numerator / denom) / inhib
        r3 = self.k[3 - n] * np.sqrt(p_co2 * p_h2) * frac

        # MTO reactions
        inhib = 1 + self.K_ch3oh * (p_ch3oh + p_h2o)
        r4 = self.k[4 - n] * p_ch3oh / inhib
        r5 = self.k[5 - n] * p_ch3oh * p_c2h4 / inhib
        r6 = self.k[6 - n] * p_ch3oh * p_c3h6 / inhib
        r7 = self.k[7 - n] * p_ch3oh * p_c4h8 / inhib
        r8_1 = self.k[8 - n] * p_c2h4 * p_h2 / inhib * self.f8["c2h4"]
        r8_2 = self.k[8 - n] * p_c3h6 * p_h2 / inhib * self.f8["c3h6"]
        r8_3 = self.k[8 - n] * p_c4h8 * p_h2 / inhib * self.f8["c4h8"]
        r9_1 = self.k[9 - n] * p_c2h4 / inhib * self.f9["c2h4"]
        r9_2 = self.k[9 - n] * p_c3h6 / inhib * self.f9["c3h6"]
        r9_3 = self.k[9 - n] * p_c4h8 / inhib * self.f9["c4h8"]
        r10_1 = self.k[10 - n] * p_c58 / inhib * self.f9["c2h4"]
        r10_2 = self.k[10 - n] * p_c58 / inhib * self.f9["c3h6"]
        r10_3 = self.k[10 - n] * p_c58 / inhib * self.f9["c4h8"]
        r11_1 = self.k[11 - n] * p_c2h4 * p_c58 / inhib * self.f9["c2h4"]
        r11_2 = self.k[11 - n] * p_c3h6 * p_c58 / inhib * self.f9["c3h6"]
        r11_3 = self.k[11 - n] * p_c4h8 * p_c58 / inhib * self.f9["c4h8"]

        return np.array(
            [
                -r1 - r2 - r3,  # co2
                -3 * r1 - r2 - 4 * r3 - r8_1 - r8_2 - r8_3,  # h2
                r1 - 9 * r4 - 2 * r5 - 3 * r6 - 4 * r7,  # ch3oh
                r1 + r2 + 2 * r3 + 9 * r4 + 2 * r5 + 3 * r6 + 4 * r7,  # h2o
                r2,  # co
                r3,  # ch4
                r4 + r5 - r8_1 - r9_1 + r10_1 - r11_1,  # c2h4
                r4 + r6 - r8_2 - r9_2 + r10_2 - r11_2,  # c3h6
                r4 + r7 - r8_3 - r9_3 + r10_3 - r11_3,  # c4h8
                r8_1,  # c2h6
                r8_2,  # c3h8
                r8_3,  # c4h10
                0.307 * r9_1
                + 0.461 * r9_2
                + 0.615 * r9_3
                - 0.307 * r10_1
                - 0.461 * r10_2
                - 0.615 * r10_3
                - r11_1
                - r11_2
                - r11_3,  # c58
                0.809 * r11_1 + 0.904 * r11_2 + r11_3,  # c9+
                0.0,  # inert
            ]
        )



import numpy as np
from numpy.typing import NDArray

import catrxneng.species as species
from .kinetic_model import KineticModel
import catrxneng.quantities as quant


class CompositeKineticModel(KineticModel):
    KINETIC_MODEL_CLASSES: list[type[KineticModel]]

    map_child_components_to_parent_components: list[list[int]]
    catalyst_fraction: NDArray

    kinetic_models: list[KineticModel]

    def __init__(
        self, catalyst_frac: NDArray, T: quant.Temperature | None = None, **kwargs
    ):
        super().__init__(T=T, **kwargs)
        self.catalyst_frac = np.asarray(catalyst_frac)
        self.kinetic_models = [KM(T=T, **kwargs) for KM in self.KINETIC_MODEL_CLASSES]
        self.map_child_components_to_parent_components = [
            [self.comp_list().index(comp) for comp in KM.comp_list()]
            for KM in self.KINETIC_MODEL_CLASSES
        ]

    @property
    def T(self):
        return self._T

    @T.setter
    def T(self, value):
        if value is not None and (not hasattr(self, "_T") or value.si != self._T.si):
            self._T = value
            try:
                for kinetic_model in self.kinetic_models:
                    kinetic_model.T = value
            except AttributeError:
                pass



import numpy as np
from numpy.typing import NDArray

from ..kinetic_model import KineticModel
import catrxneng.species as species
import catrxneng.utils.equations as eqn
import catrxneng.quantities as quant
from catrxneng.reactions import MTO


class Ghosh2022(KineticModel):
    T_REF = quant.Temperature(C=320)
    LIMITING_REACTANT = "ch3oh"
    REACTANTS = {
        "ch3oh": species.CH3OH,
        "h2": species.H2,
    }
    PRODUCTS = {
        "h2o": species.H2O,
        "c2h4": species.C2H4,
        "c3h6": species.C3H6,
        "c4h8": species.C4H8,
        "c2h6": species.C2H6,
        "c3h8": species.C3H8,
        "c4h10": species.C4H10,
        "c58": species.Species(C_ATOMS=6.5),
        "c9+": species.Species(C_ATOMS=10.5),
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {"mto": MTO}
    F8 = {
        "c2h4": 5e-2,
        "c3h6": 1.0,
        "c4h8": 2.6e-1,
    }
    F9 = {
        "c2h4": 5.5e-1,
        "c3h6": 1.0,
        "c4h8": 4.5e-1,
    }
    # ORDER = np.array([1, 2, 2, 2, 2, 1, 1, 2])
    ORDER = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    KREF_MOLSKGCATBAR = np.array(
        [5.9e-1, 7e-2, 6e-1, 5.9e-2, 8e-2, 8.3e-1, 6e-2, 8.2e-1]
    )
    EA_KJMOL = np.array([70.0, 16.0, 17.0, 69.0, 109.0, 150.0, 170.0, 25.0])
    K_ADS_REF_CH3OH = 1.3e1
    K_ADS_REF_H2O = 1.3e1
    DH_ADS_CH3OH_KJMOL = -2.0e-1
    DH_ADS_H2O_KJMOL = -2.0e-1

    def __init__(
        self,
        T: quant.Temperature | None = None,
        kref: np.typing.NDArray | None = None,
        Ea: quant.Energy | None = None,
    ):
        self.kref = kref
        if self.kref is None:
            self.kref = self.KREF_MOLSKGCATBAR
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = self.EA_KJMOL
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.Kads_ch3oh = eqn.vant_hoff_eqn(
            self.K_ADS_REF_CH3OH,
            quant.Energy(kJmol=self.DH_ADS_CH3OH_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.Kads_h2o = eqn.vant_hoff_eqn(
            self.K_ADS_REF_H2O,
            quant.Energy(kJmol=self.DH_ADS_H2O_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.k = np.array(
            [
                quant.RateConstant(
                    molskgcatbar=kref,
                    Ea=quant.Energy(kJmol=Ea),
                    Tref=self.T_REF,
                    order=order,
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.ORDER)
            ]
        )

    def calculate_rates(self, p: quant.Pressure) -> quant.ReactionRate:
        return quant.ReactionRate(
            molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list
        )

    def rate_equations(self, p_array: NDArray) -> NDArray:
        p_ch3oh = p_array[0]
        p_h2 = p_array[1]
        p_h2o = p_array[2]
        p_c2h4 = p_array[3]
        p_c3h6 = p_array[4]
        p_c4h8 = p_array[5]
        p_c58 = p_array[9]

        inhib = 1 + self.Kads_ch3oh * p_ch3oh + self.Kads_h2o * p_h2o

        n = 4
        r4 = self.k[4 - n] * p_ch3oh / inhib
        r5 = self.k[5 - n] * p_ch3oh * p_c2h4 / inhib
        r6 = self.k[5 - n] * p_ch3oh * p_c3h6 / inhib
        r7 = self.k[5 - n] * p_ch3oh * p_c4h8 / inhib
        r8_1 = self.k[8 - n] * p_c2h4 * p_h2 / inhib * self.F8["c2h4"]
        r8_2 = self.k[8 - n] * p_c3h6 * p_h2 / inhib * self.F8["c3h6"]
        r8_3 = self.k[8 - n] * p_c4h8 * p_h2 / inhib * self.F8["c4h8"]
        r9_1 = self.k[9 - n] * p_c2h4 / inhib * self.F9["c2h4"]
        r9_2 = self.k[9 - n] * p_c3h6 / inhib * self.F9["c3h6"]
        r9_3 = self.k[9 - n] * p_c4h8 / inhib * self.F9["c4h8"]
        r10_1 = self.k[10 - n] * p_c58 / inhib * self.F9["c2h4"]
        r10_2 = self.k[10 - n] * p_c58 / inhib * self.F9["c3h6"]
        r10_3 = self.k[10 - n] * p_c58 / inhib * self.F9["c4h8"]
        r11_1 = self.k[11 - n] * p_c2h4 * p_c58 / inhib * self.F9["c2h4"]
        r11_2 = self.k[11 - n] * p_c3h6 * p_c58 / inhib * self.F9["c3h6"]
        r11_3 = self.k[11 - n] * p_c4h8 * p_c58 / inhib * self.F9["c4h8"]

        return np.array(
            [
                -9 * r4 - 2 * r5 - 3 * r6 - 4 * r7,  # ch3oh
                -1 * r8_1 - r8_2 - r8_3,  # h2
                9 * r4 + 2 * r5 + 3 * r6 + 4 * r7,  # h2o
                r4 + r5 - r8_1 - r9_1 + r10_1 - r11_1,  # c2h4
                r4 + r6 - r8_2 - r9_2 + r10_2 - r11_2,  # c3h6
                r4 + r7 - r8_3 - r9_3 + r10_3 - r11_3,  # c4h8
                r8_1,  # c2h6
                r8_2,  # c3h8
                r8_3,  # c4h10
                0.307 * r9_1
                + 0.461 * r9_2
                + 0.615 * r9_3
                - 0.307 * r10_1
                - 0.461 * r10_2
                - 0.615 * r10_3
                - r11_1
                - r11_2
                - r11_3,  # c58
                0.809 * r11_1 + 0.904 * r11_2 + r11_3,  # c9+
                0.0 * r4,  # inert
            ]
        )



import numpy as np

from ..kinetic_model import KineticModel
import catrxneng.quantities as quant
import catrxneng.species as species
from catrxneng.reactions import MTO


class Lee2019(KineticModel):
    """
    kref in L/gcat/min
    conversion to normal units with moles and pressure instead of volume involves a
    calculation using the ideal gas law

    Ea in kJ/mol
    """

    TREF = quant.Temperature(C=450)
    COMPONENTS = {
        "ch4": species.CH4,
        "c2h4": species.C2H4,
        "c3h6": species.C3H6,
        "c3h8": species.C3H8,
        "c4h8": species.C4H8,
        "c4": species.Species(C_ATOMS=4),
        "c5+": species.Species(C_ATOMS=5),
        "ch3oh": species.CH3OH,
        "h2o": species.H2O,
        "inert": species.Ar,
    }
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {"mto": MTO}
    LIMITING_REACTANT = "ch3oh"

    def __init__(self, T=None, kref=None, Ea=None):
        self.kref = kref
        if self.kref is None:
            self.kref = np.array([0.16, 7.16, 7.35, 0.39, 2.26, 0.29, 0.9, 0.53, 0.36])
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = np.array([71.7, 44.15, 14.02, 3.8, 6.24, 6.24, 10.0, 13.02, 4.04])
        self.K_h2o = 6.05
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.k = np.array(
            [
                quant.RateConstant(
                    Lgcatmin=kref, Ea=quant.Energy(kJmol=Ea), Tref=self.TREF, order=1.0
                )(self.T).molhgcatbar
                for kref, Ea in zip(self.kref, self.Ea)
            ]
        )

    def calculate_rates(self, p: quant.Pressure):
        return quant.ReactionRate(
            molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list()
        )

    def rate_equations(self, p_array):
        p_ch3oh = p_array[7]
        p_h2o = p_array[8]
        p_c2h4 = p_array[1]
        p_c3h6 = p_array[2]
        P = np.sum(p_array, axis=0)

        # water mol frac calculated based on partial pressure, or set to zero if total pressure is zero
        y_h2o = np.where(P == 0, 0, p_h2o / P)
        theta_w = 1 / (1 + self.K_h2o * y_h2o)

        r = np.array(
            [
                self.k[0] * theta_w * p_ch3oh,
                self.k[1] * theta_w * p_ch3oh,
                self.k[2] * theta_w * p_ch3oh,
                self.k[3] * theta_w * p_ch3oh,
                self.k[4] * theta_w * p_ch3oh,
                self.k[5] * theta_w * p_ch3oh,
                self.k[6] * theta_w * p_ch3oh,
                self.k[7] * theta_w * p_c2h4,
                self.k[8] * theta_w * p_c3h6,
            ]
        )
        return np.array(
            [
                r[0],  # ch4
                (r[1] - r[7]) / 2,  # c2h4
                (r[2] + r[7] - r[8]) / 3,  # c3h6
                r[3] / 3,  # c3h8
                (r[4] + r[8]) / 4,  # c4h8
                r[5] / 4,  # c4
                r[6] / 5,  # c5+
                np.sum([-r[i] for i in range(7)], axis=0),  # meoh
                np.sum([r[i] for i in range(7)], axis=0),  # h2o
                0.0 * r[0],  # inert
            ]
        )



import scipy.integrate as integrate

from .reaction import Reaction
from .. import species
from .. import quantities as quant


class MeohDehydration(Reaction):
    EQUATION = "2CH₃OH ⇌ CH₃OCH₃ + H₂O"
    REACTANTS = {
        "ch3oh": species.CH3OH,
    }
    PRODUCTS = {
        "dme": species.DME,
        "h2o": species.H2O,
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    STOICH_COEFF = quant.Dimensionless(
        si=[-2.0, 1.0, 1.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "ch3oh"
    DH_RXN_298 = (
        species.DME.HF_298_GAS + species.H2O.HF_298_LIQ - 2 * species.CH3OH.HF_298_LIQ
    )
    DS_RXN_298 = (
        species.DME.S_298_GAS + species.H2O.S_298_LIQ - 2 * species.CH3OH.S_298_LIQ
    )

    @staticmethod
    def dCp_1(T_K):
        T = quant.Temperature(K=T_K)
        dCp = (
            species.DME.Cp_gas(T) + species.H2O.Cp_liq(T) - 2 * species.CH3OH.Cp_liq(T)
        )
        return dCp.JmolK

    @staticmethod
    def dCp_2(T_K):
        T = quant.Temperature(K=T_K)
        dCp = (
            species.DME.Cp_gas(T) + species.H2O.Cp_liq(T) - 2 * species.CH3OH.Cp_gas(T)
        )
        return dCp.JmolK

    @staticmethod
    def dCp_3(T_K):
        T = quant.Temperature(K=T_K)
        dCp = (
            species.DME.Cp_gas(T) + species.H2O.Cp_gas(T) - 2 * species.CH3OH.Cp_gas(T)
        )
        return dCp.JmolK

    @classmethod
    def dH_rxn(cls, T: quant.Temperature) -> quant.Energy:
        Tb_ch3oh = species.CH3OH.BOILING_TEMP.K
        Tb_h2o = species.H2O.BOILING_TEMP.K
        if T.K < Tb_ch3oh:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, T.K)[0]
            return quant.Energy(Jmol=dHr)
        elif Tb_ch3oh <= T.K < Tb_h2o:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, Tb_ch3oh)[0]
            dHr -= 2 * species.CH3OH.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_2, Tb_ch3oh, T.K)[0]
            return quant.Energy(Jmol=dHr)
        else:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, Tb_ch3oh)[0]
            dHr -= 2 * species.CH3OH.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_2, Tb_ch3oh, Tb_h2o)[0]
            dHr += species.H2O.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_3, Tb_h2o, T.K)[0]
            return quant.Energy(Jmol=dHr)

    @classmethod
    def dS_rxn(cls, T: quant.Temperature) -> quant.Entropy:
        Tb_ch3oh = species.CH3OH.BOILING_TEMP.K
        Tb_h2o = species.H2O.BOILING_TEMP.K
        integrand1 = lambda T_K: cls.dCp_1(T_K) / T_K
        integrand2 = lambda T_K: cls.dCp_2(T_K) / T_K
        integrand3 = lambda T_K: cls.dCp_3(T_K) / T_K
        if T.K < Tb_ch3oh:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, T.K)[0]
            return quant.Entropy(JmolK=dSr)
        elif Tb_ch3oh <= T.K < Tb_h2o:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_ch3oh)[0]
            dSr -= 2 * species.CH3OH.DS_VAP.JmolK
            dSr += integrate.quad(integrand2, Tb_ch3oh, T.K)[0]
            return quant.Entropy(JmolK=dSr)
        else:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_ch3oh)[0]
            dSr -= 2 * species.CH3OH.DS_VAP.JmolK
            dSr += integrate.quad(integrand2, Tb_ch3oh, Tb_h2o)[0]
            dSr += species.H2O.DS_VAP.JmolK
            dSr += integrate.quad(integrand3, Tb_h2o, T.K)[0]
            return quant.Entropy(JmolK=dSr)



# Co2ToMeoh
tags = {
    "step_name": "step_name",
    "step_num": "step_num",
    "bed_temp": "TIC02",
    "pressure": "PIC01",
    "n2_mfc": "FIC01",
    "co2_mfc": "FIC02",
    "h2_mfc": "FIC03",
    "co_mfc": "FIC04",
    "ch3oh_mfc": "FIC05",
    "h2o_mfc": "FIC06",
    "n2_gc_conc": "N2",
    "co2_gc_conc": "CO2",
    "h2_gc_conc": "H2",
    "ch3oh_gc_conc": "CH3OH",
    "dme_gc_conc": "DME",
    "co_gc_conc": "CO",
    "ch4_gc_conc": "CH4",
}



import numpy as np

from .reaction import Reaction
from ..quantities import *
from ..species import CO, H2, C2H4, H2O, Ar


class FTS(Reaction):
    COMPONENTS = {"CO": CO, "H2": H2, "C2H4": C2H4, "H2O": H2O, "inert": Ar}
    STOICH_COEFF = Dimensionless(
        si=[-2.0, -4.0, 1.0, 2.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "co"



import numpy as np
from numpy.typing import NDArray
from scipy.optimize import minimize_scalar

from catrxneng.species import Species
import catrxneng.quantities as quant


class Reaction:
    EQUATION: str
    STOICH_COEFF: quant.Dimensionless
    COMPONENTS: dict[str, type[Species]]
    DEFAULT_LIMITING_REACTANT: str

    @property
    def limiting_reactant(self):
        try:
            return self._limiting_reactant
        except AttributeError:
            return self.DEFAULT_LIMITING_REACTANT

    @limiting_reactant.setter
    def limiting_reactant(self, value):
        self._limiting_reactant = value

    @classmethod
    def comp_list(cls):
        return list(cls.COMPONENTS.keys())

    @classmethod
    def active_components(cls):
        # return {key: cls.COMPONENTS[key] for key in cls.COMPONENTS if key != "inert"}
        return {**cls.REACTANTS, **cls.PRODUCTS}

    @classmethod
    def stoich_coeff_active(cls):
        stoich_si = np.asarray(cls.STOICH_COEFF.si)
        return quant.Dimensionless(
            si=stoich_si[stoich_si != 0],
            keys=list(cls.active_components().keys()),
        )

    @classmethod
    def dH_rxn_298_gas(cls) -> quant.Energy:
        Hf_298_gas = np.array(
            [cls.COMPONENTS[key].HF_298_GAS.si for key in cls.active_components()]
        )
        dHr_298_gas = np.sum(Hf_298_gas * cls.stoich_coeff_active().si)
        return quant.Energy(si=dHr_298_gas)

    @classmethod
    def dH_rxn_gas(cls, T):
        # Hf_gas = np.array(
        #     [cls.COMPONENTS[key].Hf_gas(T).si for key in cls.active_components()]
        # )
        Hf_gas = np.array(
            [comp_cls.Hf_gas(T).si for comp_cls in cls.active_components().values()]
        )
        dHr_gas = np.sum(Hf_gas * cls.stoich_coeff_active().si)
        return quant.Energy(si=dHr_gas)

    @classmethod
    def dS_rxn_298_gas(cls) -> quant.Entropy:
        S_298_gas = np.array(
            [cls.COMPONENTS[key].S_298_GAS.si for key in cls.active_components()]
        )
        dSr_298_gas = np.sum(S_298_gas * cls.stoich_coeff_active().si)
        return quant.Entropy(si=dSr_298_gas)

    @classmethod
    def dS_rxn_gas(cls, T):
        S_gas = np.array(
            [cls.COMPONENTS[key].S_gas(T).si for key in cls.active_components()]
        )
        dSr_gas = np.sum(S_gas * cls.stoich_coeff_active().si)
        return quant.Entropy(si=dSr_gas)

    @classmethod
    def dG_rxn_298_gas(cls):
        T = quant.Temperature(K=298)
        return cls.dH_rxn_298_gas() - T * cls.dS_rxn_298_gas()

    @classmethod
    def dG_rxn(cls, T: quant.Temperature) -> quant.Energy:
        return cls.dH_rxn(T) - T * cls.dS_rxn(T)

    @classmethod
    def dG_rxn_gas(cls, T):
        return cls.dH_rxn_gas(T) - T * cls.dS_rxn_gas(T)

    @classmethod
    def Keq(cls, T: quant.Temperature) -> float:
        return np.exp(-cls.dG_rxn(T) / (quant.R * T)).si

    @classmethod
    def check_components(cls, p0, allow_component_mismatch):
        if p0.keys != list(cls.COMPONENTS.keys()) and not allow_component_mismatch:
            raise ValueError("Partial pressure keys do not match reaction components.")

    def equilibrate(
        self,
        p0: quant.Pressure,
        T: quant.Temperature,
        Keq: float | None = None,
        fug_coeff: NDArray | None = None,
        allow_component_mismatch: bool = False,
        limiting_reactant: str | None = None,
    ):
        """
        Pressure in bar
        """
        if limiting_reactant is None:
            limiting_reactant = self.DEFAULT_LIMITING_REACTANT
        self.check_components(p0, allow_component_mismatch=allow_component_mismatch)

        if allow_component_mismatch:
            p0_bar = np.array([p0[comp].bar for comp in self.comp_list()])
            p0 = quant.Pressure(bar=p0_bar.astype(float), keys=self.comp_list())

        P_bar = np.sum(p0.bar)
        initial_total_moles = 100.0
        initial_molfrac = p0.bar / P_bar
        initial_moles = initial_molfrac * initial_total_moles

        std_state_fugacity_bar = quant.Pressure(
            atm=np.ones(len(self.COMPONENTS)).astype(float)
        ).bar
        if fug_coeff is None:
            fug_coeff = np.ones(len(self.COMPONENTS)).astype(float)

        stoich_coeff = self.STOICH_COEFF.si
        if Keq is None:
            Keq = type(self).Keq(T=T)

        def objective(extent):
            moles = initial_moles + extent * stoich_coeff
            if np.any(moles < 0):
                return 1e10
            total_moles = np.sum(moles)
            molfrac = moles / total_moles
            fugacity = molfrac * fug_coeff * P_bar
            activity = fugacity / std_state_fugacity_bar
            Ka = np.prod(activity**stoich_coeff)
            if Ka <= 0:
                return 1e10
            log_error = np.log(Ka) - np.log(Keq)  # type: ignore
            return log_error * log_error

        # Calculate extent bounds from stoichiometry
        # Maximum extent limited by complete consumption of any reactant
        adj_init_mol_reactants = np.array(
            [
                mol / stoich
                for mol, stoich in zip(initial_moles, stoich_coeff)
                if stoich < 0
            ]
        )
        min_extent = 1e-5  # Small positive value to avoid numerical issues
        max_extent = np.min(-adj_init_mol_reactants) * 0.999  # Leave small margin

        # Use bounded scalar minimization - more efficient and robust for 1D problems
        result = minimize_scalar(
            objective,
            bounds=(min_extent, max_extent),
            method="bounded",
            options={"xatol": 1e-8, "maxiter": 500},
        )

        if result.success:
            self.extent = quant.Moles(si=result.x)
            initial_moles = quant.Moles(si=initial_moles, keys=self.comp_list())
            moles = initial_moles + self.extent * self.STOICH_COEFF
            self.eq_conversion = (
                initial_moles[limiting_reactant] - moles[limiting_reactant]
            ) / initial_moles[limiting_reactant]
            total_moles = quant.Moles(si=np.sum(moles.si))
            self.eq_molfrac = moles / total_moles
        else:
            raise ValueError("Optimization failed: " + result.message)



class Species:
    CLASS = "species"
    C_ATOMS = 0
    H_ATOMS = 0
    O_ATOMS = 0
    N_ATOMS = 0
    MOL_WEIGHT = 0.0
    NIST_THERMO_PARAMS: list

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def _get_thermo_params(cls, T):
        for thermo_params in cls.NIST_THERMO_PARAMS:
            if thermo_params["min_temp_K"] <= T.K <= thermo_params["max_temp_K"]:
                return thermo_params
        raise ValueError(
            f"Temperature outside range for {cls.__name__} thermodynamic parameters."
        )



from scipy.optimize import fsolve
from .reaction import Reaction
from ..species import CO, H2, CH4, H2O, Ar
from ..quantities import Dimensionless


class COMethanation(Reaction):
    def __init__(self, limiting_reactant="CO"):
        self.components = {
            "CO": CO(),
            "H2": H2(),
            "CH4": CH4(),
            "H2O": H2O(),
            "inert": Ar(),
        }
        self.stoich_coeff = Dimensionless(
            si=[-1.0, -3.0, 1.0, 2.0, 0.0], keys=list(self.components.keys())
        )
        super().__init__(limiting_reactant=limiting_reactant)



from scipy.optimize import fsolve

from ..quantities import *
from .reaction import Reaction
from ..species import CO2, H2, C2H4, H2O, Ar


class CO2FTS(Reaction):
    def __init__(self, limiting_reactant="CO2"):
        self.components = {
            "CO2": CO2(),
            "H2": H2(),
            "C2H4": C2H4(),
            "H2O": H2O(),
            "inert": Ar(),
        }
        self.stoich_coeff = Dimensionless(
            si=[-2.0, -6.0, 1.0, 4.0, 0.0], keys=list(self.components.keys())
        )
        super().__init__(limiting_reactant=limiting_reactant)



import numpy as np

from .reaction import Reaction
from .. import species
from ..quantities import *
from .rwgs import RWGS


class WGS(Reaction):
    COMPONENTS = {
        "co": species.CO,
        "h2o": species.H2O,
        "co2": species.CO2,
        "h2": species.H2,
        "inert": species.Ar,
    }
    STOICH_COEFF = Dimensionless(
        si=[-1.0, -1.0, 1.0, 1.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "co"

    @staticmethod
    def Keq(T):
        # return 1 / RWGS.Keq(T)
        return np.reciprocal(RWGS.Keq(T))



from .ghosh2022_composite import Ghosh2022Composite
from .ghosh2022_full import Ghosh2022Full
from .ghosh_lee import GhoshLee
from .ghosh_ghoshmod import GhoshGhoshMod
from .power_law_Ghosh2022_simplified import PowerLawGhosh2022Simplified



import plotly.graph_objects as go
import numpy as np
import plotly.express as px
from typing import Any
from .plotly_plot import PlotlyPlot


class BasicPlot(PlotlyPlot):

    def __init__(self, title: str | None = None, color_palette: str | None = None):
        self.fig = go.Figure()
        self.title = title
        if color_palette:
            self.colors = getattr(px.colors.qualitative, color_palette)
            self.fig.update_layout(colorway=self.colors)
        else:
            self.colors = self.COLORS
        self.right_axis = False
        self.info_text = ""

    def add_trace(
        self,
        x: np.ndarray,
        y: np.ndarray,
        name: str | None = None,
        mode: str = "lines",
        yaxis: str = "y1",
        hover_labels: list | None = None,
        color: str | None = None,
        dash: str | None = None,
        symbol: str | None = None,
    ):
        # if isinstance(y, (float, int, np.number)):
        #     xmin = np.min(x)
        #     xmax = np.max(x)
        #     x = [xmin, xmax]
        #     y = [y, y]
        hovertemplate = None
        if hover_labels:
            hovertemplate = "<b>%{text}</b><br>X: %{x}<br>Y: %{y}<extra></extra>"
        # If the caller asked for the secondary y axis, remember that so
        # render() can create it even when no ylabel2 is explicitly passed.
        if yaxis == "y2":
            self.right_axis = True

        trace_kwargs: dict[str, Any] = dict(
            x=x,
            y=y,
            mode=mode,
            yaxis=yaxis,
            name=name,
            text=hover_labels,
            hovertemplate=hovertemplate,
            showlegend=name is not None,
        )
        if mode == "lines":
            trace_kwargs["line"] = dict(dash=dash, color=color)
        if mode == "markers":
            trace_kwargs["marker"] = dict(symbol=symbol, color=color)

        trace = go.Scatter(**trace_kwargs)
        self.fig.add_trace(trace)

    def add_vertical_line(
        self,
        x: float,
        line_dash: str = "dash",
        line_color: str = "black",
        line_width: int = 2,
    ):
        """
        Adds a vertical dashed line at the specified x-value.

        Args:
            x_value (float): The x-coordinate where the vertical line will be added.
            line_dash (str): The dash style of the line (default is "dash").
            line_color (str): The color of the line (default is "black").
            line_width (int): The width of the line (default is 2).
        """
        self.fig.add_shape(
            type="line",
            x0=x,
            x1=x,
            y0=0,
            y1=1,
            xref="x",
            yref="paper",
            line=dict(
                dash=line_dash,
                color=line_color,
                width=line_width,
            ),
        )

    # def show(self):
    #     self.fig.show()

    def format(
        self,
        xlabel: str,
        ylabel: str,
        xrange: list = None,
        yrange: list = None,
        y2label: str = None,
        y2range: list = None,
        # info_text=None,
    ):
        # if xrange is None:
        #     xrange = [None, None]
        # if yrange is None:
        #     yrange = [None, None]
        # if y2range is None:
        #     y2range = [None, None]

        # if info_text:
        #     self.info_text = info_text
        top = 50
        width = 700
        if self.info_text == "":
            bottom = 50
            height = 400
        else:
            bottom = 110
            height = None

        yaxis2 = None
        # If a secondary axis label was provided, create yaxis2 with that
        # title. Also create a default right-hand axis when any trace asked
        # for `yaxis='y2'` (self.right_axis) even if no ylabel2 was passed.
        if y2label:
            yaxis2 = dict(
                title=f"<b>{y2label}</b>",
                range=y2range,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
                overlaying="y",
                side="right",
            )
        elif self.right_axis:
            # create a minimal secondary axis so traces assigned to 'y2'
            # will be rendered on the right side.
            yaxis2 = dict(
                title="",
                range=y2range,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
                overlaying="y",
                side="right",
            )
        self.fig.update_layout(
            title=dict(text=f"<b>{self.title}</b>", x=0.5),
            xaxis_title=f"<b>{xlabel}</b>",
            yaxis_title=f"<b>{ylabel}</b>",
            width=width,
            height=height,
            margin=dict(t=top, b=bottom),
            yaxis=dict(
                range=yrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
            ),
            yaxis2=yaxis2,
            xaxis=dict(
                range=xrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
            ),
            legend=dict(x=1.05, y=1, xanchor="left", yanchor="top"),
            annotations=self.plot_info_box(self.info_text),
            plot_bgcolor="white",
            paper_bgcolor="white",
            colorway=self.COLORS,
        )

    def render(
        self,
        xlabel: str,
        ylabel: str,
        xrange: list = None,
        yrange: list = None,
        ylabel2: str = None,
        yrange2: list = None,
        # info_text=None,
    ):
        self.format(
            xlabel=xlabel,
            ylabel=ylabel,
            xrange=xrange,
            yrange=yrange,
            y2label=ylabel2,
            y2range=yrange2,
        )
        self.fig.show()

    def add_info_text(self, *args):
        for line in args:
            if self.info_text:
                self.info_text += "\n" + line
            else:
                self.info_text = line



import numpy as np
from numpy.typing import NDArray

from ..kinetic_model import KineticModel
import catrxneng.utils.equations as eqn
import catrxneng.species as species
import catrxneng.quantities as quant
from catrxneng.reactions import RWGS, Co2ToMeoh


class Ghosh2021WithoutSabatier(KineticModel):
    T_REF = quant.Temperature(C=300)
    LIMITING_REACTANT = "co2"
    REACTANTS = {
        "co2": species.CO2,
        "h2": species.H2,
    }
    PRODUCTS = {
        "ch3oh": species.CH3OH,
        "h2o": species.H2O,
        "co": species.CO,
        "inert": species.Inert,
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {
        "co2_to_meoh": Co2ToMeoh,
        "rwgs": RWGS,
    }
    ORDER = np.array([2.0, 1.5, 1.0])
    STOICH_COEFF = np.array(
        [
            [-1, -3, 1, 1, 0, 0],
            [-1, -1, 0, 1, 1, 0],
        ]
    )
    KREF_MOLSKGCATBAR = np.array([6.9e-4, 1.8e-3])
    EA_KJMOL = np.array([35.7, 54.5])
    K_ADS_REF_H2 = 0.76
    K_ADS_REF_CO2 = 0.79
    DH_ADS_H2_KJMOL = -12.5
    DH_ADS_CO2_KJMOL = -25.9

    def __init__(self, site_model="single", T=None, kref=None, Ea=None):
        self.site_model = site_model
        self.kref = kref
        if self.kref is None:
            self.kref = self.KREF_MOLSKGCATBAR
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = self.EA_KJMOL
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.Kads_h2 = eqn.vant_hoff_eqn(
            self.K_ADS_REF_H2,
            quant.Energy(kJmol=self.DH_ADS_H2_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.Kads_co2 = eqn.vant_hoff_eqn(
            self.K_ADS_REF_CO2,
            quant.Energy(kJmol=self.DH_ADS_CO2_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.Keq_co2_to_meoh = Co2ToMeoh.Keq(self.T)
        self.Keq_rwgs = RWGS.Keq(self.T)
        self.Keq = np.array([self.Keq_co2_to_meoh, self.Keq_rwgs])
        self.k = np.array(
            [
                quant.RateConstant(
                    molskgcatbar=kref,
                    Ea=quant.Energy(kJmol=Ea),
                    Tref=self.T_REF,
                    order=order,
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.ORDER)
            ]
        )

    def calculate_rates(self, p: quant.Pressure) -> quant.ReactionRate:
        return quant.ReactionRate(
            molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list()
        )

    def rate_equations(self, p_array: NDArray) -> NDArray:
        """
        Calculate reaction rates from partial pressures.

        Pressure in bar
        Rates in mol/h/gcat

        Parameters
        ----------
        p_array : array-like
            Partial pressures. Can be:
            - 1D array of shape (7,) for a single point
            - 2D array of shape (7, n) for n points

        Returns
        -------
        rates : ndarray
            Reaction rates with the same shape as input.
            - 1D array of shape (7,) if input is 1D
            - 2D array of shape (7, n) if input is 2D
        """
        p_co2 = p_array[0]  # co2
        p_h2 = p_array[1]  # h2
        p_ch3oh = p_array[2]  # ch3oh
        p_h2o = p_array[3]  # h2o
        p_co = p_array[4]  # co

        base = 1 + self.K_co2 * p_co2 + np.sqrt(self.K_h2 * p_h2)
        inhib = base * base

        # Reaction 1: CO2 + 3H2 -> CH3OH + H2O (CO2-to-MeOH)
        fwd = p_co2 * p_h2 * p_h2 * p_h2
        rev = p_ch3oh * p_h2o / self.K_co2_to_meoh
        numerator = self.k[0] * (fwd - rev) / (p_h2 * p_h2)
        r1 = numerator / inhib

        # Reaction 2: CO2 + H2 -> CO + H2O (RWGS)
        fwd = p_co2 * p_h2
        rev = p_co * p_h2o / self.K_rwgs
        numerator = self.k[1] * (fwd - rev) / np.sqrt(p_h2)
        r2 = numerator / inhib

        return np.array(
            [
                -r1 - r2,  # co2
                -3 * r1 - r2,  # h2
                r1,  # ch3oh
                r1 + r2,  # h2o
                r2,  # co
                0.0 * r1,  # inert
            ]
        )



import numpy as np
from .kinetic_model import KineticModel
from catrxneng.reactions import Reaction
from catrxneng.species import Species
from .. import quantities as quant


class ReverseKineticModel(KineticModel):
    FORWARD_MODEL: type[KineticModel]
    COMPONENTS: dict[str, Species]
    REACTIONS: dict[str, Reaction]
    LIMITING_REACTANT: str
    C_ATOMS: np.ndarray

    forward_model: KineticModel

    def __init__(self, T: quant.Temperature = None, **forward_model_kwargs):
        self.forward_model = self.FORWARD_MODEL(T=T, **forward_model_kwargs)
        super().__init__(T)

    def compute_temp_dependent_constants(self) -> None:
        self.forward_model.compute_temp_dependent_constants()

    def calculate_rates(self, p: quant.Pressure) -> quant.ReactionRate:
        return -1.0 * self.forward_model.calculate_rates(p)

    def rate_equations(self, p_array):
        return -1.0 * self.forward_model.rate_equations(p_array)



import numpy as np
from numpy.typing import NDArray

from catrxneng import species
from catrxneng import reactions
from ..kinetic_model import KineticModel
import catrxneng.utils.equations as eqn
import catrxneng.quantities as quant


class Ghosh2022Simplified(KineticModel):
    LIMITING_REACTANT = "ch3oh"
    T_REF = quant.Temperature(C=320)
    REACTANTS = {
        "ch3oh": species.CH3OH,
        "h2": species.H2,
    }
    PRODUCTS = {
        "h2o": species.H2O,
        "c2h4": species.C2H4,
        "c3h6": species.C3H6,
        "c4h8": species.C4H8,
        "c5h10": species.C5H10,
        "c2h6": species.C2H6,
        "c3h8": species.C3H8,
        "c4h10": species.C4H10,
        "c5h12": species.C5H12,
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {"mto": reactions.MTO}
    ORDER = np.array([1, 1, 1, 1, 2, 2, 2, 2])
    KREF_MOLSKGCATBAR = np.array(
        [5.9e-1, 7e-2, 6e-1, 5.9e-2, 8e-2, 8.3e-1, 6e-2, 8.2e-1]
    )
    # KREF_MOLSKGCATBAR = np.array(
    #     [1000.0, 800.0, 600.0, 400.0, 0.001, 0.001, 0.001, 0.001]
    # )
    EA_KJMOL = np.array([70.0, 16.0, 17.0, 69.0, 109.0, 150.0, 170.0, 25.0])
    # EA_KJMOL = np.array(
    #     [
    #         70.0,
    #         70.0,
    #         70.0,
    #         70.0,
    #         110.0,
    #         110.0,
    #         110.0,
    #         110.0,
    #     ]
    # )
    K_ADS_REF_CH3OH = 1.3e1
    K_ADS_REF_H2O = 1.3e1
    DH_ADS_CH3OH_KJMOL = -2.0e-1
    DH_ADS_H2O_KJMOL = -2.0e-1

    def __init__(
        self,
        T: quant.Temperature | None = None,
        kref: np.typing.NDArray | None = None,
        Ea: quant.Energy | None = None,
    ):
        self.kref = kref
        if self.kref is None:
            self.kref = self.KREF_MOLSKGCATBAR
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = self.EA_KJMOL
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.Kads_ch3oh = eqn.vant_hoff_eqn(
            self.K_ADS_REF_CH3OH,
            quant.Energy(kJmol=self.DH_ADS_CH3OH_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.Kads_h2o = eqn.vant_hoff_eqn(
            self.K_ADS_REF_H2O,
            quant.Energy(kJmol=self.DH_ADS_H2O_KJMOL),
            self.T,
            self.T_REF,
        ).si
        self.k = np.array(
            [
                quant.RateConstant(
                    molskgcatbar=kref,
                    Ea=quant.Energy(kJmol=Ea),
                    Tref=self.T_REF,
                    order=order,
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.ORDER)
            ]
        )

    def calculate_rates(self, p: quant.Pressure) -> quant.ReactionRate:
        return quant.ReactionRate(
            molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list()
        )

    def rate_equations(self, p_array: NDArray) -> NDArray:
        p_ch3oh = p_array[0]
        p_h2 = p_array[1]
        p_h2o = p_array[2]
        p_c2h4 = p_array[3]
        p_c3h6 = p_array[4]
        p_c4h8 = p_array[5]
        p_c5h10 = p_array[6]
        inhib = 1 + self.Kads_ch3oh * p_ch3oh + self.Kads_h2o * p_h2o

        r_c2h4 = self.k[0] * p_ch3oh / inhib
        r_c3h6 = self.k[1] * p_ch3oh / inhib
        r_c4h8 = self.k[2] * p_ch3oh / inhib
        r_c5h10 = self.k[3] * p_ch3oh / inhib
        r_c2h6 = self.k[4] * p_c2h4 * p_h2 / inhib
        r_c3h8 = self.k[5] * p_c3h6 * p_h2 / inhib
        r_c4h10 = self.k[6] * p_c4h8 * p_h2 / inhib
        r_c5h12 = self.k[7] * p_c5h10 * p_h2 / inhib

        return np.array(
            [
                -2 * r_c2h4 - 3 * r_c3h6 - 4 * r_c4h8 - 5 * r_c5h10,  # ch3oh
                -r_c2h4 - r_c3h6 - r_c4h8 - r_c5h10,  # h2
                2 * r_c2h4 + 3 * r_c3h6 + 4 * r_c4h8 + 5 * r_c5h10,  # h2o
                r_c2h4 - r_c2h6,  # c2h4
                r_c3h6 - r_c3h8,  # c3h6
                r_c4h8 - r_c4h10,  # c4h8
                r_c5h10 - r_c5h12,  # c5h10
                r_c2h6,  # c2h6
                r_c3h8,  # c3h8
                r_c4h10,  # c4h10
                r_c5h12,  # c5h12
                0.0 * r_c2h4,  # inert
            ]
        )



from .quantity import Quantity
from catrxneng import utils


class TimeDelta(Quantity):

    @property
    def sec(self):
        return self.si

    @sec.setter
    def sec(self, value):
        self.si = utils.self.to_float(value)

    @property
    def min(self):
        return utils.divide(self.si, 60)

    @min.setter
    def min(self, value):
        self.si = value * 60

    @property
    def hr(self):
        return utils.divide(self.si, 3600)

    @hr.setter
    def hr(self, value):
        self.si = value * 3600



import scipy.integrate as integrate

from .reaction import Reaction
from ..quantities import *
from ..species import *


class MTO(Reaction):
    COMPONENTS = {
        "ch3oh": CH3OH,
        "c2h4": C2H4,
        "h2o": H2O,
        "inert": Ar,
    }
    STOICH_COEFF = Dimensionless(si=[-2.0, 1.0, 2.0, 0.0], keys=list(COMPONENTS.keys()))
    DEFAULT_LIMITING_REACTANT = "ch3oh"

    DH_RXN_298 = C2H4.HF_298_GAS + 2 * H2O.HF_298_LIQ - 2 * CH3OH.HF_298_LIQ
    DS_RXN_298 = C2H4.S_298_GAS + 2 * H2O.S_298_LIQ - 2 * CH3OH.S_298_LIQ
    DH_RXN_298_GAS = C2H4.HF_298_GAS + 2 * H2O.HF_298_GAS - 2 * CH3OH.HF_298_GAS
    DS_RXN_298_GAS = C2H4.S_298_GAS + 2 * H2O.S_298_GAS - 2 * CH3OH.S_298_GAS

    # @staticmethod
    # def dCp_1(T_K):
    #     T = Temperature(K=T_K)
    #     dCp = CH3OH.Cp_liq(T) + H2O.Cp_liq(T) - CO2.Cp_gas(T) - 3 * H2.Cp_gas(T)
    #     return dCp.JmolK

    # @staticmethod
    # def dCp_2(T_K):
    #     T = Temperature(K=T_K)
    #     dCp = CH3OH.Cp_gas(T) + H2O.Cp_liq(T) - CO2.Cp_gas(T) - 3 * H2.Cp_gas(T)
    #     return dCp.JmolK

    # @staticmethod
    # def dCp_3(T_K):
    #     T = Temperature(K=T_K)
    #     dCp = CH3OH.Cp_gas(T) + H2O.Cp_gas(T) - CO2.Cp_gas(T) - 3 * H2.Cp_gas(T)
    #     return dCp.JmolK

    # @classmethod
    # def dH_rxn(cls, T):
    #     Tb_ch3oh = CH3OH.BOILING_TEMP.K
    #     Tb_h2o = H2O.BOILING_TEMP.K
    #     if T.K < Tb_ch3oh:
    #         dHr = cls.DH_RXN_298.Jmol
    #         dHr += integrate.quad(cls.dCp_1, 298, T.K)[0]
    #         return Energy(Jmol=dHr)
    #     if Tb_ch3oh <= T.K < Tb_h2o:
    #         dHr = cls.DH_RXN_298.Jmol
    #         dHr += integrate.quad(cls.dCp_1, 298, Tb_ch3oh)[0]
    #         dHr += CH3OH.DH_VAP.Jmol
    #         dHr += integrate.quad(cls.dCp_2, Tb_ch3oh, T.K)[0]
    #         return Energy(Jmol=dHr)
    #     if T.K >= Tb_h2o:
    #         dHr = cls.DH_RXN_298.Jmol
    #         dHr += integrate.quad(cls.dCp_1, 298, Tb_ch3oh)[0]
    #         dHr += CH3OH.DH_VAP.Jmol
    #         dHr += integrate.quad(cls.dCp_2, Tb_ch3oh, Tb_h2o)[0]
    #         dHr += H2O.DH_VAP.Jmol
    #         dHr += integrate.quad(cls.dCp_3, Tb_h2o, T.K)[0]
    #         return Energy(Jmol=dHr)

    # @classmethod
    # def dS_rxn(cls, T):
    #     Tb_ch3oh = CH3OH.BOILING_TEMP.K
    #     Tb_h2o = H2O.BOILING_TEMP.K
    #     integrand1 = lambda T_K: cls.dCp_1(T_K) / T_K
    #     integrand2 = lambda T_K: cls.dCp_2(T_K) / T_K
    #     integrand3 = lambda T_K: cls.dCp_3(T_K) / T_K
    #     if T.K < Tb_ch3oh:
    #         dSr = cls.DS_RXN_298.JmolK
    #         dSr += integrate.quad(integrand1, 298, T.K)[0]
    #         return Entropy(JmolK=dSr)
    #     if Tb_ch3oh <= T.K < Tb_h2o:
    #         dSr = cls.DS_RXN_298.JmolK
    #         dSr += integrate.quad(integrand1, 298, Tb_ch3oh)[0]
    #         dSr += CH3OH.DS_VAP.JmolK
    #         dSr += integrate.quad(integrand2, Tb_ch3oh, T.K)[0]
    #         return Entropy(JmolK=dSr)
    #     if T.K >= Tb_h2o:
    #         dSr = cls.DS_RXN_298.JmolK
    #         dSr += integrate.quad(integrand1, 298, Tb_ch3oh)[0]
    #         dSr += CH3OH.DS_VAP.JmolK
    #         dSr += integrate.quad(integrand2, Tb_ch3oh, Tb_h2o)[0]
    #         dSr += H2O.DS_VAP.JmolK
    #         dSr += integrate.quad(integrand3, Tb_h2o, T.K)[0]
    #         return Entropy(JmolK=dSr)



import plotly.express as px
import plotly.graph_objects as go


class PlotlyPlot:
    fig: go.Figure
    # COLORS = px.colors.qualitative.Plotly
    # COLORS = [
    #     "#EB3437",
    #     "#714BDB",
    #     "#34B13F",
    #     "#E951B6",
    #     "#F0752E",
    #     "#76AAAC",
    #     "#0A0101",
    #     "#80490B",
    #     "#98ADF1",
    #     "#F8F411",
    # ]
    COLORS = [
        "red",
        "blue",
        "limegreen",
        "magenta",
        "#F0752E",  # orange
        "gray",
        "black",
        "lightblue",
        "brown",
        "yellow",
        "cyan",
    ]

    @staticmethod
    def plot_info_box(text):
        if text is None:
            return None
        text = text.replace("\\n", "\n")
        text = text.strip("\n")
        text = text.replace("\n", "<br>")
        annotations = [
            dict(
                x=0.5,
                y=-0.35,
                xref="paper",
                yref="paper",
                text=text,
                showarrow=False,
                font=dict(size=12, style="italic", weight="bold"),
            )
        ]
        return annotations



import pandas as pd
import numpy as np
from catrxneng.reactors.pfr import PFR


class Step:

    @property
    def lab_notebook_id(self):
        return f"{self.expt.lab_notebook_id}-s{self.steady_state_step_num}"

    def __init__(self, expt, step_name, step_num, start, end, reactor: PFR):
        from .expt import Expt
        self.expt: Expt = expt
        self.step_name = step_name
        self.step_num = step_num
        self.start = start
        self.end = end
        self.reactor: PFR = reactor

    def simulate(self, dt_sec, std_dev=None):
        self.std_dev = std_dev
        if std_dev is None:
            self.std_dev = {"temp": 0.3, "pressure": 0.005, "mfc": 0.005, "gc": 0.005}
        df = pd.DataFrame()
        df["timestamp"] = np.arange(self.start.UET, self.end.UET + dt_sec, dt_sec)
        self.num_points = df["timestamp"].size
        df["step_name"] = self.step_name
        df["step_num"] = self.step_num
        self.reactor.solve()
        df["bed_temp"] = self.reactor.T.C + np.random.normal(
            loc=0, scale=self.std_dev["temp"], size=self.num_points
        )
        df["pressure"] = self.reactor.P.bar * np.random.normal(
            loc=1, scale=self.std_dev["pressure"], size=self.num_points
        )

        for component in self.reactor.y.keys:
            df[f"{component}_gc_conc"] = self.reactor.y[-1][
                component
            ].pct * np.random.normal(
                loc=1, scale=self.std_dev["gc"], size=self.num_points
            )

        df["total_gc_conc"] = df.filter(like="gc_conc").sum(axis=1)
        self.time_series_data = df

        # add additional simulate logic in child class simulate method after calling super().simulate

    # def populate_inlet_partial_pressures(self):
    #     df = self.time_series_data
    #     for component in self.reactor.kinetic_model_class.comp_list():
    #         df[f"p_{component}_bar"] = df[f"{component}_gc_conc"] / 100 * df["pressure"]



from .quantity import Quantity
from .energy import Energy
from .temperature import Temperature
from catrxneng import utils


class HeatCapacity(Quantity):

    @property
    def JmolK(self):
        return self.si

    @JmolK.setter
    def JmolK(self, value):
        self.si = self.to_float(value)

    @property
    def kJmolK(self):
        return self.si / 1000

    @kJmolK.setter
    def kJmolK(self, value):
        self.si = self.to_float(value * 1000)

    # def __mul__(self, other):
    #     keys = self.get_keys(self, other)
    #     if isinstance(other, Temperature):
    #         si = self.si * other.si
    #         return Energy(si=si, keys=keys)
    #     else:
    #         return super().__mul__(other)

    # def __rmul__(self, other):
    #     keys = self.get_keys(self, other)
    #     if isinstance(other, Temperature):
    #         si = other.si * self.si
    #         return Energy(si=si, keys=keys)
    #     else:
    #         return super().__rmul__(other)



import os
import pickle
import pandas as pd
from pathlib import Path
from typing import Any, Optional
from catrxneng.plots import PlotlyPlot
from catrxneng.simulate import Expt
from catrxneng.kinetic_models.kinetic_model import KineticModel
from catrxneng.reactors import Reactor


class Notebook:
    expt: Optional[Expt]
    km: Optional[KineticModel]
    reactor: Optional[Reactor]
    df: Optional[pd.DataFrame]
    data: Optional[dict[str, Any]]

    def __init__(self, script_path: str, descrip: str | None = None) -> None:
        self.script_path = script_path
        self.script_name = Path(script_path).stem
        self.script_dir = Path(script_path).parent
        self.path = self.script_path.replace(".py", ".html")
        self.html = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>{self.script_name}</title>
            <meta charset="utf-8" /> 
            <style>
                body {{
                    text-align: center;
                    margin: 0 auto;
                    max-width: 1200px;
                    padding: 20px;
                }}
                h1 {{
                    margin-bottom: 30px;
                }}
                .plotly-graph-div {{
                    margin: 0 auto;
                }}
            </style>
        </head>
        <body>
            {f'<h1>{self.script_name}</h1>'}
        </body>
        </html>
        """
        if descrip:
            self.add_text_to_html(descrip)

    def add_plot_to_html(self, plot: PlotlyPlot) -> None:
        include_plotlyjs = "inline" if "cdn.plot.ly" not in self.html else False
        plot_html = plot.fig.to_html(include_plotlyjs=include_plotlyjs)
        # head_content = plot_html.split("<head>")[1].split("</head>")[0]
        body_content = plot_html.split("<body>")[1].split("</body>")[0]
        # if html == "":
        #     html = initialize_html(script_name)
        self.html = self.html.replace(
            "</body>",
            f"<br><br>\n{body_content}\n<br>\n</body>",
        )

    @staticmethod
    def add_line_to_text(text: str, *args, color: str | None = None) -> str:
        for line in args:
            if color:
                line = f'<span style="color: {color};">{line}</span>'
            text += line + "<br>"
        return text

    def add_text_to_html(self, text: str) -> None:
        centered_text = f'<div style="text-align: center;">{text}</div>'
        self.html = self.html.replace(
            "</body>", f" <br>\n {centered_text}\n <br>\n</body>"
        )

    def add_table_to_html(self, table: str) -> None:
        centered_table = (
            f'<div style="display: flex; justify-content: center;">{table}</div>'
        )
        self.html = self.html.replace(
            "</body>", f" <br>\n {centered_table}\n <br>\n</body>"
        )

    def save_html(self) -> None:
        with open(self.path, "w") as f:
            f.write(self.html)

    def save_dataframe(self, var_name: str, df: pd.DataFrame) -> None:
        filename = f"{self.script_name}_{var_name}.parquet"
        full_path = os.path.join(self.script_dir, filename)
        df.to_parquet(full_path)

    def save_pickle(self, var_name: str, data: Any) -> None:
        filename = f"{self.script_name}_{var_name}.pkl"
        full_path = os.path.join(self.script_dir, filename)
        with open(full_path, "wb") as f:
            pickle.dump(data, f)

    def load_dataframe(self, var_name: str) -> Optional[pd.DataFrame]:
        parquet_filename = f"{self.script_name}_{var_name}.parquet"
        parquet_full_path = os.path.join(self.script_dir, parquet_filename)
        try:
            return pd.read_parquet(parquet_full_path)
        except (FileNotFoundError, Exception):
            return None

    def load_pickle(self, var_name: str) -> Optional[Any]:
        pkl_filename = f"{self.script_name}_{var_name}.pkl"
        pkl_full_path = os.path.join(self.script_dir, pkl_filename)
        try:
            with open(pkl_full_path, "rb") as f:
                return pickle.load(f)
        except (FileNotFoundError, pickle.UnpicklingError, EOFError):
            return None



from .reaction import Reaction
from .. import species
from ..quantities import Dimensionless


class AmmoniaSynthesis(Reaction):
    COMPONENTS = {
        "n2": species.N2,
        "h2": species.H2,
        "nh3": species.NH3,
        "inert": species.Ar,
    }
    # STOICH_COEFF = Unitless(si=[-0.5, -1.5, 1, 0], keys=list(COMPONENTS.keys()))
    STOICH_COEFF = Dimensionless(si=[-1, -3, 2, 0], keys=list(COMPONENTS.keys()))
    DEFAULT_LIMITING_REACTANT = "n2"

    @classmethod
    def dH_rxn(cls, T):
        return cls.dH_rxn_gas(T)

    @classmethod
    def dS_rxn(cls, T):
        return cls.dS_rxn_gas(T)



import scipy.integrate as integrate

from .reaction import Reaction
from ..quantities import *
from ..species import *


class RWGS(Reaction):
    COMPONENTS = {
        "co2": CO2,
        "h2": H2,
        "co": CO,
        "h2o": H2O,
        "inert": Ar,
    }
    STOICH_COEFF = Dimensionless(
        si=[-1.0, -1.0, 1.0, 1.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "co2"
    DH_RXN_298 = CO.HF_298_GAS + H2O.HF_298_LIQ - CO2.HF_298_GAS - H2.HF_298_GAS
    DH_RXN_298_GAS = CO.HF_298_GAS + H2O.HF_298_GAS - CO2.HF_298_GAS - H2.HF_298_GAS
    DS_RXN_298 = CO.S_298_GAS + H2O.S_298_LIQ - CO2.S_298_GAS - H2.S_298_GAS

    @staticmethod
    def dCp_1(T_K):
        T = Temperature(K=T_K)
        dCp = CO.Cp_gas(T) + H2O.Cp_liq(T) - CO2.Cp_gas(T) - H2.Cp_gas(T)
        return dCp.JmolK

    @staticmethod
    def dCp_2(T_K):
        T = Temperature(K=T_K)
        dCp = CO.Cp_gas(T) + H2O.Cp_gas(T) - CO2.Cp_gas(T) - H2.Cp_gas(T)
        return dCp.JmolK

    @classmethod
    def dH_rxn(cls, T):
        Tb_h2o = H2O.BOILING_TEMP.K
        if T.K < Tb_h2o:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, T.K)[0]
            return Energy(Jmol=dHr)
        if T.K >= Tb_h2o:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, Tb_h2o)[0]
            dHr += H2O.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_2, Tb_h2o, T.K)[0]
            return Energy(Jmol=dHr)

    @classmethod
    def dS_rxn(cls, T):
        Tb_h2o = H2O.BOILING_TEMP.K
        integrand1 = lambda T_K: cls.dCp_1(T_K) / T_K
        integrand2 = lambda T_K: cls.dCp_2(T_K) / T_K
        if T.K < Tb_h2o:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, T.K)[0]
            return Entropy(JmolK=dSr)
        if T.K >= Tb_h2o:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_h2o)[0]
            dSr += H2O.DS_VAP.JmolK
            dSr += integrate.quad(integrand2, Tb_h2o, T.K)[0]
            return Entropy(JmolK=dSr)



from .quantity import Quantity
from catrxneng.utils import *


class ArbitraryUnits(Quantity):

    def __init__(self, value, units, units_pwr):
        self.value = value
        self.units = units
        self.units_pwr = units_pwr



import scipy.integrate as integrate, numpy as np

from .reaction import Reaction
from ..quantities import Dimensionless, Temperature, Energy, Entropy
from ..species import CO2, H2, CH4, H2O, Ar


class Sabatier(Reaction):
    COMPONENTS = {
        "co2": CO2,
        "h2": H2,
        "ch4": CH4,
        "h2o": H2O,
        "inert": Ar,
    }

    STOICH_COEFF = Dimensionless(
        si=[-1.0, -4.0, 1.0, 2.0, 0.0], keys=list(COMPONENTS.keys())
    )

    DH_RXN_298 = (
        CH4.HF_298_GAS + 2 * H2O.HF_298_LIQ - CO2.HF_298_GAS - 4 * H2.HF_298_GAS
    )

    DH_RXN_298_GAS = (
        CH4.HF_298_GAS + 2 * H2O.HF_298_GAS - CO2.HF_298_GAS - 4 * H2.HF_298_GAS
    )

    DS_RXN_298 = (
        CH4().S_298_GAS + 2 * H2O().S_298_LIQ - CO2().S_298_GAS - 4 * H2().S_298_GAS
    )

    DEFAULT_LIMITING_REACTANT = "co2"

    # def __init__(self, limiting_reactant="co2"):
    # self.components = {
    #     "co2": CO2(),
    #     "h2": H2(),
    #     "ch4": CH4(),
    #     "h2o": H2O(),
    #     "inert": Ar(),
    # }
    # self.stoich_coeff = Unitless(
    #     si=[-1.0, -4.0, 1.0, 2.0, 0.0], keys=list(self.components.keys())
    # )
    # super().__init__(limiting_reactant=limiting_reactant)

    @staticmethod
    def dCp_1(T_K):
        T = Temperature(K=T_K)
        dCp = CH4.Cp_gas(T) + 2 * H2O.Cp_liq(T) - CO2.Cp_gas(T) - 4 * H2.Cp_gas(T)
        return dCp.JmolK

    @staticmethod
    def dCp_2(T_K):
        T = Temperature(K=T_K)
        dCp = CH4.Cp_gas(T) + 2 * H2O.Cp_gas(T) - CO2.Cp_gas(T) - 4 * H2.Cp_gas(T)
        return dCp.JmolK

    @classmethod
    def dH_rxn(cls, T):
        Tb_h2o = H2O.BOILING_TEMP.K
        if T.K < Tb_h2o:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, T.K)[0]
            return Energy(Jmol=dHr)
        if T.K >= Tb_h2o:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, Tb_h2o)[0]
            dHr += H2O.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_2, Tb_h2o, T.K)[0]
            return Energy(Jmol=dHr)

    @classmethod
    def dS_rxn(cls, T):
        Tb_h2o = H2O.BOILING_TEMP.K
        integrand1 = lambda T_K: cls.dCp_1(T_K) / T_K
        integrand2 = lambda T_K: cls.dCp_2(T_K) / T_K
        if T.K < Tb_h2o:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, T.K)[0]
            return Entropy(JmolK=dSr)
        if T.K >= Tb_h2o:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_h2o)[0]
            dSr += H2O.DS_VAP.JmolK
            dSr += integrate.quad(integrand2, Tb_h2o, T.K)[0]
            return Entropy(JmolK=dSr)



from .expt import Expt
from .step import Step
from .co2_fts_step import Co2FtsStep
from .meoh_to_co2_step import MeohToCo2Step
from .tandem_step import TandemStep
from .co2_to_meoh_step import Co2ToMeohStep